﻿using System;
using System.Collections.Generic;
using EZ_B;

namespace Dynamixel {

  // http://www.trossenrobotics.com/images/productdownloads/AX-12(English).pdf

  public class DynamixelV1 {

    protected internal DynamixelV1() {

    }

    byte[] createDynamixelCommand(byte id, byte[] buffer) {

      if (buffer.Length > 255)
        throw new Exception("AX-12 Command cannot be longer than 255 bytes");

      List<byte> bList      = new List<byte>();
      byte       dataLength = (byte)(buffer.Length + 1);

      bList.Add(0xff);
      bList.Add(0xff);
      bList.Add(id);
      bList.Add(dataLength);

      int checksum = id;

      checksum += dataLength;

      for (int x = 0; x < buffer.Length; x++) {

        checksum += buffer[x];

        bList.Add(buffer[x]);
      }

      bList.Add((byte)(0xff - (checksum % 256)));

      return bList.ToArray();
    }

    public enum BAUD_RATES {
      BAUD_1000000 = 1,
      BAUD_500000  = 3,
      BAUD_400000  = 4,
      BAUD_250000  = 7,
      BAUD_200000  = 9,
      BAUD_115200  = 16,
      BAUD_57600   = 34,
      BAUD_19200   = 103,
      BAUD_9600    = 207
    }

    /// <summary>
    /// Change Baud Rate
    /// </summary>
    public byte[] ChangeBaudRate(byte id, BAUD_RATES baud) {

      List<byte> buffer = new List<byte>();

      buffer.Add(3);  // Write data instruction

      buffer.Add(0x04); // Baud Rate

      buffer.Add((byte)baud);

      return createDynamixelCommand(id, buffer.ToArray());
    }

    /// <summary>
    /// Change Baud Rate
    /// </summary>
    public byte[] ChangeBaudRate(Servo.ServoPortEnum servo, BAUD_RATES baud) {

      byte id = (byte)(servo - Servo.ServoPortEnum.V0);

      return ChangeBaudRate(id, baud);
    }

    /// <summary>
    /// Factory Default
    /// </summary>
    public byte[] ResetToFactoryDefault(byte id) {

      List<byte> buffer = new List<byte>();

      buffer.Add(0x06);  // Reset instruction

      return createDynamixelCommand(id, buffer.ToArray());
    }

    /// <summary>
    /// Factory Default
    /// </summary>
    public byte[] ResetToFactoryDefault(Servo.ServoPortEnum servo) {

      byte id = (byte)(servo - Servo.ServoPortEnum.V0);

      return ResetToFactoryDefault(id);
    }

    /// <summary>
    /// Return the data packet that will release a servo with the specified ID
    /// </summary>
    public byte[] ReleaseServo(byte id) {

      List<byte> buffer = new List<byte>();

      buffer.Add(3);  // Write data instruction

      buffer.Add(24); // Torque Enable

      buffer.Add(0x00);

      return createDynamixelCommand(id, buffer.ToArray());
    }

    /// <summary>
    /// Returns the data packet that will  move a servo with the specified id to the position
    /// </summary>
    public byte[] MoveServoCmd(byte id, int position) {

      if (position < 0)
        position = 0;

      List<byte> buffer = new List<byte>();

      buffer.Add(3);  // Write data instruction

      buffer.Add(30); // Goal Position

      buffer.AddRange(BitConverter.GetBytes((UInt16)position));

      return createDynamixelCommand(id, buffer.ToArray());
    }

    /// <summary>
    /// Return a packet that will set the speed of the servo with the id to the speed
    /// </summary>
    public byte[] ServoSpeed(byte id, int speed) {

      if (speed < 0)
        speed = 0;

      List<byte> buffer = new List<byte>();

      buffer.Add(3);  // Write data instruction

      buffer.Add(34); // Goal Position

      buffer.AddRange(BitConverter.GetBytes((UInt16)speed));

      return createDynamixelCommand(id, buffer.ToArray());
    }

    /// <summary>
    /// Change the LED status of the dynamixel servo
    /// </summary>
    public byte[] LED(Servo.ServoPortEnum servo, bool status) {

      if (servo < Servo.ServoPortEnum.V0 || servo > Servo.ServoPortEnum.V99)
        throw new Exception("Servo out of range for Dynamixel");

      byte id = (byte)(servo - Servo.ServoPortEnum.V0);

      return LED(id, status);
    }

    /// <summary>
    /// Change the LED status of the specified servo
    /// </summary>
    public byte[] LED(byte id, bool status) {

      List<byte> buffer = new List<byte>();

      buffer.Add(3);  // Write data instruction
      buffer.Add(25); // LED
      buffer.Add(status ? (byte)1 : (byte)0);

      return createDynamixelCommand(id, buffer.ToArray());
    }

    public byte[] DisableAllAlarms(byte id) {

      List<byte> buffer = new List<byte>();

      buffer.Add(3);  // Write data instruction
      buffer.Add(0x12);
      buffer.Add(254);

      return createDynamixelCommand(id, buffer.ToArray());
    }

    public byte[] DisableAllAlarms(Servo.ServoPortEnum servo) {

      byte id = (byte)(servo - Servo.ServoPortEnum.V0);

      return DisableAllAlarms(id);
    }

    public byte[] DisableAllAlarms() {

      return DisableAllAlarms(0xfe);
    }

    public byte[] GetDisableStatusPacketCmd(byte id) {

      List<byte> buffer = new List<byte>();

      buffer.Add(3);  // Write data instruction
      buffer.Add(0x10);
      buffer.Add(0);

      return createDynamixelCommand(id, buffer.ToArray());
    }

    public byte[] DisableStatusPacket(Servo.ServoPortEnum servo) {

      byte id = (byte)(servo - Servo.ServoPortEnum.V0);

      return GetDisableStatusPacketCmd(id);
    }

    public byte[] DisableStatusPacket() {

      return GetDisableStatusPacketCmd(0xfe);
    }

    public byte[] ChangeID(Servo.ServoPortEnum source, Servo.ServoPortEnum destination) {

      byte       souceID       = (byte)(source - Servo.ServoPortEnum.V0);
      byte       destinationID = (byte)(destination - Servo.ServoPortEnum.V0);
      List<byte> buffer        = new List<byte>();

      buffer.Add(3);  // Write data instruction
      buffer.Add(3);  // Change ID
      buffer.Add(destinationID);

      return createDynamixelCommand(souceID, buffer.ToArray());
    }
  }
}
